# A tweek file consists of lines represented by Tweek::Entry
#
# All of the entries in the file have a line number and optional comment
# other attributes depend on the specific type.
#
# Lifecycle: An entry is created either directly or parsed from an input line,
# then the 'condfail' method is called to see if it's relevant, if so
# the actual values are filled in according to the section. Then the 'mismatch'
# method can be used to see if it matched. Finally the 'generate' method
# produces a human-readable output
#
require 'ostruct'

class Tweek::Entry < OpenStruct

  # Not all entries are settable: record if this entry was used to set/reset a value or not
  attr_accessor :mode

  # The type of this entry:
  # :literal        With line attributes - single lines
  # :section        With section type and list attributes
  # :parameter      With parameter value and conditional attributes
  #
  attr_reader :type

  def initialize section, type, h
    raise "Invalid Entry type #{type}" unless [:literal, :parameter, :section].include? type
    super h
    @section = section
    @type = type
    @mode = :query
    @condfail = nil
    @mismatch = nil
  end

  # Parse an input line
  #
  def self.parse section, line

    line.chomp!
    comment = line.slice!(/#.*$/)
    comment.slice!(/^#\s*/) if comment
    entry = case
            when line.empty?
              Tweek::Entry.new( section, :literal, :line => line, :comment => comment )

            when /^\s*([A-Z0-9]+)\s*(.*?)\s*$/ =~ line # We've hit a new section
              Tweek::Entry.new( section, :section, :section_type => $1.to_sym, :list => $2,
                                :comment => comment )

            when /^\s*(.+?)\s*=\s*(.*?)\s*(\bif\b(.*))?$/ =~ line # A parameter line
              param = $1
              value = $2
              raise "Missing condition after 'if'" if $4 and $4.strip.empty?
              cond = $4.nil? ? nil : $4.strip
              was = /\s*\[was\s+(.*)\]/.match(comment)
              if was
                comment = was.pre_match+was.post_match
                was = was.captures.first
              end
              expected = /\s*\[expected\s+(.*)\]/.match(comment)
              if expected
                comment = expected.pre_match+was.post_match
                expected = expected.captures.first
              end
              comment = nil if comment and comment.empty?
              Tweek::Entry.new( section, :parameter, :param => param, :value => value,
                                :cond => cond, :comment => comment,
                                :was => was, :expected => expected )
            else
              raise "Unrecognized line"
            end
    return entry
  end

  # Check if any associated condition is true or false
  #
  def condfail
    @condfail ||= @section.twf.condfail(cond)
  end

  # Check if the entry matches reality
  #
  def mismatch
    raise "Value or Actual missing! Can't check for mismatch" if value.nil? or actual.nil?
    @mismatch ||= !(value === actual)
  end

  # Generate an output line for this entry
  #
  def generate

    outline = nil

    case self.type

    when :literal
      outline = line
      outline += "# #{comment}" if comment

    when :section
      if section_type.nil?
        outline = nil
      else
        name = list_element || list
        outline = "#{section_type} #{name} #{comment}"
      end

    when :parameter
      condout = cond.nil? ? "": " if #{cond}"
      if condfail
        body = "#{param} = #{value}#{condout.colorize(:red)}"
        note = nil
      else
        swap! if mode != :query
        if mismatch
          body = "#{param} = #{actual.colorize(:red)}#{condout.colorize(:green)}"
          note = " [#{mode != :query ? 'was' : 'expected'} #{value.colorize(:yellow)}]"
        else
          body = "#{param} = #{actual}#{condout.colorize(:green)}"
          note = nil
        end
      end
      if comment.nil?
        outline = (note ? (body + " #" + note) : body)
      else
        outline = (note ? (body + " # " + comment + note) : (body + " # " + comment))
      end
    end
    return outline

  end

  # Method to toggle the actual and expected values, used for :query, :set, :reset
  def swap!
      self.actual, self.value = self.value, self.actual
  end

end
