require 'rubygems'
require 'tweek'
require 'tweek/section'
require 'tweek/entry'
require 'colorize'

class Tweek::File

  attr_reader :distro, :distro_version, :kernel_version, :mflag, :oflag, :mismatches, :skips

  def initialize( distro, distro_version, kernel_version, mflag = :query, oflag = 0)
    @distrov = distro
    @dv = Gem::Version.new(distro_version)
    @kv = Gem::Version.new(kernel_version)
    @distro_version = distro_version.bold
    @kernel_version = kernel_version.bold
    @distro = distro.bold
    @nparams = 0
    @mismatches = 0
    @errors = []
    @skips = 0
    @warns = []
    @generated = []
    @mflag = mflag
    @oflag = oflag

  end

  # Called for every entry, generates output as required
  #
  def generate entry

    line = entry.generate

    case entry.type

    when :literal
      @generated.push line if @oflag > 1

    when :section
      @generated.push line unless entry.section_type.nil? or @oflag < 2

    when :parameter
      @nparams += 1
      if entry.condfail
        @generated.push line if @oflag > 1
        STDERR.print "s".colorize(:yellow) if @oflag == 1
      else
        @generated.push line if @oflag == 3 or ((@oflag == 2) and entry.mismatch)
        if entry.mismatch
          @mismatches += 1
          STDERR.print "F".colorize(:red) if @oflag == 1
        else
          STDERR.print ".".colorize(:green) if @oflag == 1
        end
      end
    end
  end

  def error msg, lineno = 0
    line = (lineno == 0) ? msg : sprintf("%3d: #{msg}", lineno)
    @errors.push line unless @errors.first == line
  end

  def warn msg, lineno = 0
    line = (lineno == 0) ? msg : sprintf("%3d: #{msg}", lineno)
    @warns.push line unless @warns.first == line
  end

  def messages
    return @errors.size > 0 ? @errors.join("\n") : ""
  end

  def generated
    return @generated.size > 0 ? @generated.join("\n") : ""
  end

  def condfail cond
    return false unless cond
    cond.strip!
    failures = 0
    reqs = cond.split(/\s*,\s*/)
    reqs.each do |req|
      raise ArgumentError.new("entry has invalid condition variable: #{req}") unless /^[kvd]/ =~ req
      var = $&
      ver = $'
      ok =  case var
            when 'k'
              Gem::Requirement.new(ver).satisfied_by?(@kv)
            when 'v'
              Gem::Requirement.new(ver).satisfied_by?(@dv)
            when 'd'
              op = ver.slice!(/^[<>~!=]+/)
              begin
                eval "@distrov #{op} #{ver}"
              rescue Exception => e
                raise ArgumentError.new("entry has condition error: #{e.message}")
              end
            end
      failures += 1 unless ok
    end
    skipped = (failures > 0)
    @skips += 1 if skipped
    return skipped
  end

  # Read the entire file and split into sections
  #
  def read_sections handle
    section = Tweek::Section.initial self
    while line = handle.gets
      line.chomp!
      if (line =~ /^=begin/)..(line =~ /^=end/)
        entry = Tweek::Entry.new(section, :literal, :line => line, :lineno => handle.lineno )
        section.push entry
        next
      end

      begin
        entry = Tweek::Entry.parse(section, line)
        entry.lineno = handle.lineno
        case entry.type

        when :literal
          section.push entry

        when :section
          section.process
          section = Tweek::Section.new(self, entry)

        when :parameter
          section.push entry

        else
          error "#{handle.lineno}: Unrecognized line: #{line}"
        end
      rescue
        error "#{handle.lineno}: #{$!.message}"
      end
    end
    section.process
  end

  def results
    unless @oflag == 0
      STDERR.puts "" if @oflag == 1
      STDERR.puts "Mode: #{@mflag.to_s.bold} Distro: #{@distro} Version: #{@distro_version} Kernel: #{@kernel_version}"
      STDERR.puts "#{@nparams} parameters checked, #{@skips} conditions not met, #{@mismatches} mismatches, #{@warns.size} warnings, #{@errors.size} errors"
      STDERR.puts "#{@errors.join("\n")}" unless @errors.empty?
      STDERR.puts "#{@warns.join("\n")}" unless @warns.empty?
      if oflag > 1
        @generated.unshift "# Generated at #{Time.now.strftime("%c %Z")}"
        @generated.unshift "# Mode: #{@mflag} Distro: #{@distro} Version: #{@distro_version} Kernel: #{@kernel_version}"
      end
      @generated.each { |l| STDOUT.puts l}
    end
    return @mismatches
  end
end
