# A section is an array of lines, each line is an Entry
# The section line itself is stored in a separate Entry object
#
class Tweek::Section < Array

  attr_reader :section_entry, :twf

  def initialize twf, entry
    @twf = twf
    @section_entry = entry
  end

  # Return an initial section used to hold anything (such as comments) prior to
  # the first real section
  def self.initial twf
    Tweek::Section.new(twf, Tweek::Entry.new(self, :section, :section_type => :initial) )
  end

  # Helper method to easily iterate through each actionable entry in the section.
  #
  def each_entry &block
    self.each do |entry|
      case entry.type
      when :literal
        @twf.generate entry
      when :parameter
        begin
          if entry.condfail
            @twf.generate entry
          else
            yield entry if block_given?
          end
        rescue Exception => e
          @twf.error e.message, entry.lineno
          STDERR.puts "#{e.message}\n#{e.backtrace.join("\n")}" if ENV['DEBUG_TWEEK']
        end
      else
        raise "Unknown entry type: #{entry.type}"
      end
    end
  end

  def process

    return if self.empty?

    case section_entry.section_type

    when :initial
      self.each do |entry|
        case entry.type
        when :literal
          @twf.generate entry
        else
          @twf.warn "Parameter line ignored outside section", entry.lineno
        end
      end

    when :BLKDEV # BLOCK DEVICE PARAMETERS
      @twf.warn "Set/Reset is not currently implemented for #{section_entry.section_type} section" if [:set, :reset].include? @twf.mflag
      Dir.chdir('/dev')
      devices = Dir.glob(section_entry.list.split(' ').map{|dev| dev.gsub(/^\/dev\//,'')})
      @twf.warn("Block device check skipped: no devices found", section_entry.lineno) if devices.empty?
      devices.each do |blk|
        section_entry.list_element = blk
        @twf.generate section_entry
        each_entry do |entry|
          entry.actual = Tweek.blk_param(blk, entry.param)
          if entry.param == 'scheduler'
            entry.actual = entry.actual.slice(/\[(.*?)\]/,1) unless entry.actual == 'none'
          end
          @twf.generate entry
        end
      end

    when :KERNEL # KERNEL PARAMETERS
      @twf.warn "Set/Reset is not currently implemented for #{section_entry.section_type} section" if [:set, :reset].include? @twf.mflag
      @twf.generate section_entry
      each_entry do |entry|
        actual = /\b(#{entry.param})(=\S+)?\b/.match(Tweek.bootline)
        case
        when (entry.value == 'true' and actual)
          entry.actual = actual[2].nil? ? 'true' : actual[2][1..-1]
        when (entry.value == 'true' and actual.nil?)
          entry.actual = 'false'
        when (entry.value == 'false' and actual)
          entry.actual = 'true'
        when (entry.value == 'false' and actual.nil?)
          entry.actual = 'false'
        else
         entry.actual = actual.nil? ? 'false' : actual[2][1..-1]
        end
        @twf.generate entry
      end

    when :CLOCKSOURCE
      @twf.warn "Set/Reset is not currently implemented for #{section_entry.section_type} section" if [:set, :reset].include? @twf.mflag
      @twf.generate section_entry
      each_entry do |entry|
        entry.actual = Tweek.clocksource(entry.param)
        @twf.generate entry
      end

    when :SYSCTL # SYSCTL SETTINGS
      @twf.generate section_entry
      each_entry do |entry|
        file = entry.param.gsub(/\./,'/')
        entry.actual = Tweek.sysctl(file)
        case @twf.mflag
        when :set
          entry.value = Tweek.set_sysctl(file, entry.value)
          entry.mode = :set
        when :reset
          if entry.was
            if entry.actual == entry.value
              entry.value = Tweek.set_sysctl(file, entry.was)
              entry.mode = :reset
            else
              @twf.warn "Current value of #{entry.param.bold} does not match, cannot reset", entry.lineno
            end
          else
            @twf.warn "Previous value of #{entry.param.bold} not found, cannot reset", entry.lineno
          end
        end
        @twf.generate entry
      end

    when :NET  # NETWORK DEVICES
      @twf.warn "Set/Reset is not currently implemented for #{section_entry.section_type} section" if [:set, :reset].include? @twf.mflag
      devices = section_entry.list.split(' ').map{|dev| dev.gsub(/^\/dev\//,'')}
      @twf.warn("Network device check skipped: no devices found", section_entry.lineno) if devices.empty?

      devices.each do |netdev|
        unless Tweek.net_dev_exist? netdev
          @twf.warn("NET device #{netdev} not found", section_entry.lineno)
          next
        end
        section_entry.list_element = netdev
        @twf.generate section_entry
        each_entry do |entry|

	  case entry.param
	  when 'driver'
            entry.actual = Tweek.net_driver(netdev)
            @twf.generate entry

	  when 'irqs'
            balpid = Tweek.balance_pid
            if entry.value == 'balance'
              entry.actual = balpid.zero? ? 'IRQ balance daemon not running' : 'balance'
            else
              if balpid.zero?
                handle_irqs(netdev, entry)
              else
                entry.actual = 'IRQ balance daemon running'
                @twf.generate entry
              end
            end

	  else # treat the parameter as a name
            entry.actual = Tweek.net_param(netdev, entry.param)
            @twf.generate entry
          end

        end
      end

    when :EXT4 # EXT4 filesystems info via `dumpe2fs -h /dev/#{device}`
      @twf.warn "Set/Reset is not currently implemented for #{section_entry.section_type} section" if [:set, :reset].include? @twf.mflag
      mounted = Tweek.mounted
      mounts = Dir.glob(section_entry.list.split(' '))
      @twf.warn("EXT4 check skipped: no mountpoints found", section_entry.lineno) if mounts.empty?
      mounts.each do |mount|
        unless this = mounted.select{|m| (mount == m[0] or mount == m[1]) and m[2] == 'ext4'}.first
          @twf.warn "EXT4 path #{mount} is not mounted or is not mounted as ext4", section_entry.lineno
          next
        end
        section_entry.list_element = mount
        @twf.generate section_entry
        device = this[0].gsub('/dev/','')
        optstring = File.open("/proc/fs/ext4/#{device}/options",&:read)
        options = Hash[optstring.scan(/([^=\s]+)=([^=\s,]+)/)] # options a=b
        optstring.split(/\n/).reject{|o| o.index('=')}.each{|o| options[o] = 'true'}
        each_entry do |entry|
          entry.actual = options[entry.param] || "<not found>"
          @twf.generate entry
        end
      end

    when :XFS # XFS filesystems info  via `xfs_info`
      @twf.warn "Set/Reset is not currently implemented for #{section_entry.section_type} section" if [:set, :reset].include? @twf.mflag
      # Dynamically via: /proc/fs/xfs/... ?
      mounts = section_entry.list.split(' ')
      @twf.warn("XFS device check skipped: no mountpoints found", section_entry.lineno) if mounts.empty?
      mounts.each do |mount|
        xfsinfo = Tweek.xfsinfo(mount)
        unless xfsinfo
          @twf.warn "No XFS filesystem at #{mount}", section_entry.lineno
          next
        end
        section_entry.list_element = mount
        @twf.generate section_entry
        data = Hash[xfsinfo.slice!(/^meta-data.*(?=^naming)/m).scan(/([^=\s]+)=([^=\s,]+)/)]
        naming = Hash[xfsinfo.slice!(/^naming.*(?=^log)/m).scan(/([^=\s]+)=([^=\s,]+)/)]
        log = Hash[xfsinfo.slice!(/^log.*(?=^realtime)/m).scan(/([^=\s]+)=([^=\s,]+)/)]
        realtime = Hash[xfsinfo.slice!(/^realtime.*$/m).scan(/([^=\s]+)=([^=\s,]+)/)]
        xfsparms = { 'data' => data, 'naming' => naming, 'log' => log, 'realtime' => realtime }

        each_entry do |entry|
          parameter = entry.param.split('.',2)
          entry.actual = xfsparms[parameter[0]][parameter[1]] rescue "<invalid name>"
          @twf.generate entry
        end
      end

    else
      @twf.error "Unknown type #{section_entry.section_type}", section_entry.lineno
    end

  end

  def handle_irqs netdev, entry
    irqs = Tweek.irqs(netdev)

    if irqs.empty?
      @twf.warn "#{netdev} IRQ check skipped: none found", entry.lineno
      entry.actual = 'null'
      return
    end

    ncores = Tweek.ncores
    mismatches = 0
    expected_rps = []
    expected_rss = []
    irqs.each_with_index do |irq, ix|
      case entry.value
      when 'null','none','ignore'
        next

      when 'aws'
        case ncores
        when 8..64
          expected_rps << ("00000000," * 3) + "0000ff00"
          expected_rss << ("00000000," * 3) + sprintf("%08x", 1<<ix)
        when 1..8
          expected_rps << "0000"
          expected_rss << sprintf("%04x", 1<<ix)
        else
          @twf.warn "#{netdev} IRQ strategy '#{entry.value}' can't handle #{ncores} cores", entry.lineno
          return
        end

      when 'pin'
        case ncores
        when 1..8 # not a lot of choice
          expected_rps << "0000"
          expected_rss << "0001"

        when 8..32 # RPS on cores 10-15 and 18-31, RSS tied to 0-7
          case ix
          when 0..7
            expected_rps << ("00000000," * 3) + "fffefe00"
            expected_rss << ("00000000," * 3) + sprintf("%08x", 1<<(ix+1))
          else
            @twf.warn "#{netdev} IRQ strategy '#{entry.value}' can't handle IRQ#{ix}", entry.lineno
            return
          end

        when 32..64 # Split the work across the two nodes
          case ix
          when 0..3 # alloc on node 0 (cpus 0-3)
            expected_rps << ("00000000," * 2) + "0000ffff," + "0000ffe0"
            expected_rss << ("00000000," * 3) + sprintf("%08x", 1<<(ix+1))
          when 4..7 # alloc on node 1 (
            expected_rps << ("00000000," * 2) + "ffff0000," + "ffe00000"
            expected_rss << ("00000000," * 3) + sprintf("%08x", 1<<(ix + 9))
          else
            @twf.warn "#{netdev} IRQ strategy '#{entry.value}' can't handle IRQ#{ix}", entry.lineno
            return
          end

        else
          @twf.error "Can't handle #{ncores} cores in IRQ strategy '#{entry.value}'", entry.lineno
          return
        end

      else
        @twf.error "Unrecognized #{netdev} IRQ strategy '#{entry.value}'", entry.lineno
        return
      end
    end
    rss = [] # sadly map.with_index doesn't work in 1.8.7, no other compatible way
    irqs.each_with_index do |irq, ix|
      rss << Tweek::Entry.new(self, :parameter, :param => "#{netdev}_irq_#{irq}_rss",
                              :actual => Tweek.net_rss(irq),
                              :value => expected_rss[ix])
    end
    rps = []
    irqs.each_with_index do |irq, ix|
      rps << Tweek::Entry.new(self, :parameter, :param => "#{netdev}_irq_rx-#{ix}_rps",
                              :actual => Tweek.net_rps(netdev, ix),
                              :value => expected_rps[ix])
    end
    mismatches = rss.reduce(0){|sum, r| sum += 1 if r.mismatch}
    mismatches += rps.reduce(0){|sum, r| sum += 1 if r.mismatch}
    entry.actual = 'explicit' if mismatches > 0
    @twf.generate entry
    (rss+rps).each {|e| @twf.generate e}
    return
  end

end

